
import enum

NAMESPACE_FIELD_XMLNS = "xmlns"
NAMESPACE_FIELD_XSI = "xsi"
NAMESPACE_FIELD_ATTRIB_CONTENT = "{0}:{1}".format(NAMESPACE_FIELD_XMLNS, NAMESPACE_FIELD_XSI)
TEST_TICKET_ID = "test"
SCRIPTED_CONDITION_TRUE = "<response><condition_result>true</condition_result></response>"
SCRIPTED_CONDITION_FALSE = "<response><condition_result>false</condition_result></response>"

SERVICE_OBJECT_TYPE_PREDEFINED = "PREDEFINED"
SERVICE_OBJECT_TYPE_PROTOCOL = "PROTOCOL"
SERVICE_OBJECT_TYPE_APPLICATION_IDENTITY = "APPLICATION_IDENTITY"
TYPE_ANY = "ANY"
TYPE_INTERNET = "INTERNET"
TYPE_ATTRIB = "type"
TYPE_DNS = "DNS"
TYPE_IP = "IP"
TYPE_OBJECT = "Object"
TYPE_HOST = "HOST"
TYPE_NETWORK = "NETWORK"
TYPE_OTHER = "OTHER"
TYPE_RANGE = "RANGE"
TYPE_LDAP_ENTITY = "LDAP_ENTITY"
XML_ATTRIB_IDENTIFIER = "_attribs"
XML_TAG_IDENTIFIER = "_xml_tag"
XSI_NAMESPACE_URL = "http://www.w3.org/2001/XMLSchema-instance"


class Attributes:
    ACL__BINDING = "acl__binding"
    BLOCK_CELL_VIOLATION = "blocked_cell_violation"
    BLOCK_ALL_CELL_VIOLATION = "blocked_all_cell_violation"
    BLOCK_ONLY_MATRIX_CELL_VIOLATION = "blocked_only_matrix_cell_violation"
    CLOUD_SECURITY_GROUP_NETWORK_OBJECT = "cloud_security_group"
    CONCRETE = "concrete"
    CUSTOM = "custom"
    DEVICE_NETWORK = "device_network"
    DEVICE_SERVICE = "device_service"
    DNS = "dns"
    DOMAIN_NETWORK_OBJECT = "domain_network_object"
    FIELD_TYPE_APPROVE_REJECT = "approve_reject"
    FIELD_TYPE_CHECKBOX = "checkbox"
    FIELD_TYPE_DATE = "date"
    FIELD_TYPE_DROP_DOWN_LIST = "drop_down_list"
    FIELD_TYPE_HYPERLINK = "hyperlink"
    FIELD_TYPE_MANAGER = "manager"
    FIELD_TYPE_MULTIPLE_SELECTION = "multiple_selection"
    FIELD_TYPE_MULTI_ACCESS_REQUEST = "multi_access_request"
    FIELD_TYPE_MULTI_GROUP_CHANGE = "multi_group_change"
    FIELD_TYPE_MULTI_HYPERLINK = "multi_hyperlink"
    FIELD_TYPE_MULTI_NETWORK_OBJECT = "multi_network_object"
    FIELD_TYPE_MULTI_SERVICE = "multi_service"
    FIELD_TYPE_MULTI_SERVER_DECOMMISSION_REQUEST = "multi_server_decommission_request"
    FIELD_TYPE_MULTI_TARGET = "multi_target"
    FIELD_TYPE_MULTI_TEXT = "multi_text_field"
    FIELD_TYPE_MULTI_TEXT_AREA = "multi_text_area"
    FIELD_TYPE_RULE_DECOMMISSION = "rule_decommission"
    FIELD_TYPE_TEXT = "text_field"
    FIELD_TYPE_TEXT_AREA = "text_area"
    FIELD_TYPE_TIME = "time"
    GROUP = "group"
    HREF = "href"
    HOST_NETWORK_OBJECT = "host_network_object"
    HOST_NETWORK_OBJECT_WITH_INTERFACES = "host_network_object_with_interfaces"
    ICMP_SERVICE = "icmp_service"
    INTERFACE_TYPE = "interfaceDTO"
    INTERNET = "Internet"
    INTERNET_NETWORK_OBJECT = "internet_network_object"
    INSTALL_ON_NETWORK_OBJECT = "install_on_network_object"
    IP_SERVICE = "ip_service"
    NETWORK_OBJECT_GROUP = "network_object_group"
    NETWORK_OBJECT_TYPE_BASIC = "basicNetworkObjectDTO"
    NETWORK_OBJECT_TYPE_GROUP = "networkObjectGroupDTO"
    NETWORK_OBJECT_TYPE_HOST = "hostNetworkObjectDTO"
    NETWORK_OBJECT_TYPE_HOST_WITH_INTERFACES = "hostNetworkObjectWithInterfacesDTO"
    NETWORK_OBJECT_TYPE_RANGE = "rangeNetworkObjectDTO"
    NETWORK_OBJECT_TYPE_SUBNET = "subnetNetworkObjectDTO"
    NETWORK_OBJECT_TYPE_INTERNET = "internetNetworkObjectDTO"
    NETWORK_OBJECT_TYPE_CLOUD = "cloudSecurityGroupDTO"
    NETWORK_OBJECT_TYPE_VIRTUAL_SERVER = "networkObjectVirtualServerDTO"
    NETWORK_OBJECT_TYPE_VM_INSTANCE = "vmInstanceDTO"
    POLICY__BINDING = "policy__binding"
    POOL_MEMBER = "poolMember"
    POOL_MEMBERS = "poolMembers"
    PREDEFINED = "predefined"
    RANGE_NETWORK = "range_network"
    RANGE_NETWORK_OBJECT = "range_network_object"
    RANGE_SERVICE = "range_service"
    RESTRICTED_CELL_VIOLATION = 'restricted_cell_violation'
    SECURITY_ZONE_MATRIX = "securityZoneMatrixDTO"
    SECURITY_REQUIREMENT_TYPE_MATRIX = "matrixRequirementDTO"
    SECURITY_REQUIREMENT_TYPE_PCI = "pciRequirementDTO"
    SERVICE_GROUP = "service_group"
    SERVICE_TYPE_GROUP = "serviceGroupDTO"
    SERVICE_TYPE_SINGLE = "singleServiceDTO"
    SLIM_RULE_WITH_META_DATA = "slimRuleWithMetadataDTO"
    SUBNET = "subnet"
    SUBNET_NETWORK_OBJECT = "subnet_network_object"
    TRANSPORT_SERVICE = "transport_service"
    TYPE = ":type"
    VIOLATION_ANY_NETWORK_OBJECT = "any_network_object"
    VIOLATION_ANY_SERVICE = "any_service"
    VIOLATION_GROUP_NETWORK_OBJECT = "group_member_network_object"
    VIOLATION_IP_NETWORK_OBJECT = "ip_network_object"
    VIOLATION_GROUP_MEMBER_SERVICE_OBJECT = "group_member_service_object"
    VIOLATION_RANGE_NETWORK_OBJECT = "range_network_object"
    VIOLATION_SINGLE_NETWORK_OBJECT = "single_network_object"
    VIOLATION_SINGLE_SERVICE = "single_service"
    VIOLATION_SINGLE_SERVICE_OBJECT = "single_service_object"
    XSI_NAMESPACE_TYPE = "{{{0}}}type".format(XSI_NAMESPACE_URL)
    XSI_TYPE = "{0}:type".format(NAMESPACE_FIELD_XSI)
    ZONE = "zone"
    ZONE__BINDING = "zone__binding"


class Elements:
    CREATE_DATE = "createDate"
    ID = "id"
    TICKET_INFO = "ticket_info"
    UPDATE_DATE = "updateDate"
    CURRENT_STAGE = "current_stage"
    OPEN_REQUEST_STAGE = "open_request_stage"
    COMPLETION_DATA = "completion_data"
    ACCEPTINGRULESDTO = "acceptingRulesDTO"
    ACCESS_REQUEST = "access_request"
    ACCESS_REQUESTS = "access_requests"
    ACCESS_REQUEST_ID = "access_request_id"
    ACCESS_REQUEST_VERIFIER_RESULT = "access_request_verifier_result"
    ACCESS_TYPE = "access_type"
    ACL = "acl"
    ACL_NAME = "acl_name"
    ACTION = "action"
    ADDITIONAL_INFO = "additional_info"
    ADDITIONAL_PARAMETER = "additional_parameter"
    ADDITIONAL_PARAMETERS = "additional_parameters"
    ADDRESS = "address"
    ADDRESS_BOOK = "address_book"
    ADMIN = "admin"
    ADMIN_DOMAIN = "admin_domain"
    ADMINISTRATOR = "administrator"
    ALLOWED_SERVICE = "allowed_service"
    ALLOWED_SERVICES = "allowed_services"
    ANCESTOR_MANAGEMENT_ID = "ancestor_management_id"
    ANCESTOR_MANAGEMENT_NAME = "ancestor_management_name"
    ANCESTOR_MANAGEMENT_REVISION_ID = "ancestor_management_revision_id"
    ANY_ZONE = "any_zone"
    APP_NAME = "app_name"
    APPLICATION_NAME = "application_name"
    APP_OWNER = "app_owner"
    APPLICATION = "application"
    APPLICATION_ID = "application_id"
    APPLICATION_DETAILS = "application_details"
    APPLICATION_INTERFACE = "application_interface"
    APPLICATION_INTERFACES = "application_interfaces"
    APPLICATION_INTERFACE_ID = "application_interface_id"
    APPLICATION_PACK = "application_pack"
    APPLICATIONS = "applications"
    APPROVED = "approved"
    APPROVED_BY = "approved_by"
    ASSIGNED = "ASSIGNED"
    ASSIGNEE = "assignee"
    ASSIGNMENT = "assignment"
    ASSIGNEE_ID = "assignee_id"
    AUDITLOG = "auditLog"
    AUTHENTICATION_METHOD = "authentication_method"
    AUTHORIZATIONSTATUS = "authorizationStatus"
    AUTOMATIC = "automatic"
    AUTONAT = "autoNat"
    AVAILABILITY_ZONE = "availability_zone"
    BINDING = "binding"
    BINDING_NAME = "binding_name"
    BINDING_SUGGESTION = "binding_suggestion"
    BINDING_UID = "binding_uid"
    BINDINGS = "bindings"
    BINDING_AND_RULES = "binding_and_rules"
    BINDINGS_AND_RULES = "bindings_and_rules"
    BLOCKINGRULESDTO = "blockingRulesDTO"
    BLOCKSTART = "blockStart"
    BLOCKEND = "blockEnd"
    BLOCKED_SERVICE = "blocked_service"
    BLOCKED_SERVICES = "blocked_services"
    BNPP_REGION_REQUEST = "bnpp_region_request"
    BUSINESSOWNEREMAIL = "businessOwnerEmail"
    BUSINESSOWNERNAME = "businessOwnerName"
    BUSINESS_OWNERS = "business_owners"
    CHANGE_ACTION = "change_action"
    CHANGE_AUTHORIZATION = "change_authorization"
    CHANGE_AUTHORIZATION_BINDING = "change_authorization_binding"
    CHANGE_AUTHORIZATION_BINDINGS = "change_authorization_bindings"
    CHANGE_ID = "change_id"
    CHANGE_IMPLEMENTATION_STATUS = "change_implementation_status"
    CIDR = "cidr"
    CLASS_NAME = "class_name"
    CLEANUP = "cleanup"
    CLEANUPS = "cleanups"
    CLEANUP_SET = "cleanup_set"
    CLOUD_SECURITY_GROUP = "cloud_security_group"
    CLOUD_SECURITY_GROUPS = "cloud_security_groups"
    CODE = "code"
    COLOR = "color"
    COMMENT = "comment"
    COMMENTS = "comments"
    COMMUNITY = "community"
    COMMUNITIES = "communities"
    USE_TOPOLOGY = "use_topology"
    COMPLIANCE_POLICIES = "compliance_policies"
    COMPLIANCE_POLICY = "compliance_policy"
    COMPLIANCE_RULE = "compliance_rule"
    COMPLIANCE_RULES = "compliance_rules"
    CONNECTION = "connection"
    CONNECTION_EXTENDED = "connection_extended"
    CONNECTIONS_EXTENDED = "connections_extended"
    CONNECTIONS = "connections"
    CONNECTION_TO_APPLICATION = "connection_to_application"
    CONNECTION_TO_APPLICATIONS = "connection_to_applications"
    CONNECTIONS_TO_APPLICATIONS = "connections_to_applications"
    CONNECTION_TO_APPLICATION_PACK = "connection_to_application_pack"
    CONNECTION_TO_APPLICATION_PACKS = "connection_to_application_packs"
    CONNECTED_SERVERS = "connected_servers"
    CONTENT = "content"
    COUNT = "count"
    CP_INSPECT_STREAMING_NAME = "cp_inspect_streaming_name"
    CP_UID = 'cp_uid'
    CP_PROTOTYPE_NAME = "cp_prototype_name"
    CREATED = "created"
    CREATED_BY = "created_by"
    CREATION_DATE = "creation_date"
    CURRENT_STEP = "current_step"
    CUSTOMER = "customer"
    CUSTOMER_ID = "customer_id"
    CUSTOMERS = "customers"
    DATE = "date"
    DCR_PRODUCT = "dcr_product"
    DCR_TEST_CONCRETE = "dcr_test_concrete"
    DCR_TEST_GROUP = "dcr_test_group"
    DCR_TEST = "dcr_test"
    DCR_TESTS = "dcr_tests"
    DECOMMISSIONED = "decommissioned"
    DEFAULT = "default"
    DESCRIPTION = "description"
    DESIGNER_COMMAND = "designer_command"
    DESIGNER_COMMANDS = "designer_commands"
    DESIGNER_RESULT = "designer_result"
    DESIGNER_RESULTS = "designer_results"
    DESTINATION = "destination"
    DESTINATION_DOMAIN = "destination_domain"
    DESTINATIONS = "destinations"
    DESTINATIONS_IN_ZONE = "destinations_in_zone"
    DEST_NETWORKS_NEGATED = "dest_networks_negated"
    DESTINATIONSERVICES = "destinationServices"
    DEST_NETWORK_COLLECTION = "dest_network_collection"
    DESTNETWORKS = "destinationNetworks"
    DESTNEGATED = "destNegated"
    DEST_SERVICES_NEGATED = "dest_services_negated"
    DEVICE = "device"
    DEVICE_NAME = "name"
    DEVICE_INFO = "device_info"
    DEVICE_TYPE = "device_type"
    DEVICES = "devices"
    DEVICES_AND_BINDINGS = "devices_and_bindings"
    DEVICE_AND_BINDINGS = "device_and_bindings"
    DEVICE_CONFIG = "device_config"
    DEVICE_ID = "device_id"
    DEVICE_IDS = "device_ids"
    DEVICE_SUGGESTION = "device_suggestion"
    DIRECTION = "direction"
    DISABLED = "disabled"
    DISPLAYNAME = "displayName"
    DISPLAY_NAME = "display_name"
    DM_INLINE_MEMBRES = "DM_INLINE_members"
    DNSADDRESS = "dnsAddress"
    DNS_IP_ADDRESSES = 'dns_ip_addresses'
    DST = "dst"
    DOCUMENTATION = "documentation"
    RULE_DOCUMENTATION = "rule_documentation"
    DOMAIN = "domain"
    DOMAINS = "domains"
    DOMAIN_ID = "domain_id"
    DOMAIN_NAME = "domain_name"
    DST_NAT_METHOD = "dstNatMethod"
    DST_NETWORK = "dst_network"
    DST_NETWORKS = "dst_networks"
    DST_SERVICE = "dst_service"
    DST_SERVICES = "dst_services"
    EDIT_MODE = "edit_mode"
    EDITORS = "editors"
    EDITOR = "editor"
    EMAIL = "email"
    ENABLE_NET_4_TO_NET_6 = "enable_net4tonet6"
    ENABLE_ROUTE_LOOKUP = "enable_route_lookup"
    ENGRESS_INTERFACE = "egress_interface"
    ENFORCEDON = "enforcedOn"
    EXCLUDE_ANY = "exclude_any"
    EXCLUSION = "exclusion"
    EXEMPTED_TRAFFIC = "exempted_traffic"
    EXEMPTED_TRAFFIC_LIST = "exempted_traffic_list"
    EXPIRATION_DATE = "expiration_date"
    EXPIRATION_FIELD_NAME = "expiration_field_name"
    EXPIREDATE = "expireDate"
    EXPRESSION = "expression"
    EXTERNAL = "external"
    FIELD = "field"
    FIELDS = "fields"
    FIRST_IP = "first_ip"
    FIRST_NAME = "first_name"
    FROM = "from"
    FLOW = "flow"
    FLOW_DESTINATION = 'flow_destination'
    FLOW_DESTINATION_VIOLATIONS = "flow_destination_violations"
    FLOW_DESTINATIONS = 'flow_destinations'
    FLOW_SOURCE = "flow_source"
    FLOW_SOURCE_VIOLATIONS = "flow_source_violations"
    FLOW_SOURCES = "flow_sources"
    FROM_DOMAIN = "from_domain"
    FROM_ZONE = "from_zone"
    F5_DEVICE_NAME = "f5_device_name"
    GENERIC_DEVICES = "generic_devices"
    GLOBAL = "global"
    GROUP = "group"
    GROUP_CHANGE = "group_change"
    GROUP_ID = "group_id"
    GROUP_MEMBER = "group_member"
    GROUP_MEMBER_PATH = "group_member_path"
    GROUPPERMISSION = "groupPermission"
    GROUPPERMISSIONS = "groupPermissions"
    GROUPID = "groupId"
    GUICLIENT = "guiClient"
    HANDLE_IMPLICIT_CLEANUP_RULE = "handled_by_implicit_cleanup_rule"
    HOST_NAME = "host_name"
    HYPERLINK = "hyperlink"
    ID = "id"
    IMPLEMENTATION_PERCENTAGE_THRESHOLD = "implementation_percentage_threshold"
    IMPLEMENTATION_STATUS = "implementation_status"
    IMPLEMENTING_RULE = "implementing_rule"
    IMPLEMENTING_RULES = "implementing_rules"
    IMPLEMENTS_ACCESS_REQUESTS = "implements_access_requests"
    IMPLICIT = "implicit"
    INCOMING_INTERFACE_NAME = "incoming_interface_name"
    INCOMINGINTERFACES = "incomingInterfaces"
    INCOMINGVRF = "incomingVrf"
    INDOMAINELEMENTID = "inDomainElementId"
    INPUT = "input"
    INSTALL = "install"
    INSTALLABLE_TARGET = "installable_target"
    INSTALL_ON = "install_on"
    INSTALL_ONS = "install_ons"
    INSTALLED_ON_MODULE = "installed_on_module"
    INSTANCE_ID = "instance_id"
    INSTANCES_TOTAL = "instances_total"
    INSTRUCTION = "instruction"
    INSTRUCTION_TYPE = "instruction_type"
    INSTRUCTIONS = "instructions"
    INTERFACE = "interface"
    INTERFACES = "interfaces"
    INTERFACE_FOR_NETWORK_OBJECT = "interface_for_network_object"
    INTERFACE_CONNECTION = "interface_connection"
    INTERFACE_CONNECTIONS = "interface_connections"
    INTERFACE_IP = "interface_ip"
    INTERFACE_IPS = "interface_ips"
    INTERFACE_NAME = "interface_name"
    INTERNET_REFERRAL_OBJECT = "internet_referral_object"
    INTERVAL = "interval"
    IP = "ip"
    IPSECLIST = "ipsecList"
    IP_ADDRESS = "ip_address"
    IP_TYPE = "ip_type"
    IS_ANY = "is_any"
    ISACTIVE = "isActive"
    ISCUSTOM = "isCustom"
    ISDEFAULT = "isDefault"
    ISDISABLED = "is_disabled"
    ISMANDATORY = "isMandatory"
    IS_PUBLISHED = "is_published"
    ITG = "itg"
    ITG_ID = "itg_id"
    KEY = "key"
    LABEL = "label"
    LABELS = "labels"
    LAST_HIT = "last_hit"
    LAST_IP = "last_ip"
    LAST_MODIFIED = "last_modified"
    LAST_NAME = "last_name"
    LDAP_ENTITY_ID = "ldap_entity_id"
    LDAP_ENTITY_DN = "ldap_entity_dn"
    LDAP_ENTITY_NAME = "ldap_entity_name"
    LDAPDN = "ldapDn"
    LDAP_CONFIGURATION = "ldap_configuration"
    LEGACY_RULE = "legacy_rule"
    LEVEL = "level"
    LINK = "link"
    MANAGEMENTID = "managementId"
    MANAGEMENT_DOMAIN = "management_domain"
    MANAGEMENT_ID = "management_id"
    MANAGEMENT_IP = "management_ip"
    MANAGEMENT_NAME = "management_name"
    MANAGEMENT_TYPE = "management_type"
    MASK = "mask"
    MATCH_FOR_ANY = "match_for_any"
    MATCH_RULE = "match_rule"
    MATRIX_CELL_VIOLATION = "matrix_cell_violation"
    MAX = "max"
    MAX_PROTOCOL = "max_protocol"
    MAX_VALUE_SOURCE = "max_value_source"
    MAXIP = "maxIp"
    MAX_IP = "max_ip"
    MAX_ICMP_TYPE = "max_icmp_type"
    MAX_PORT = "max_port"
    MEMBER = "member"
    MEMBER_OF = "member_of"
    MEMBERS = "members"
    MESSAGE = "message"
    MGMT_ANY = "mgmt_any"
    MGMT = "mgmt"
    MGMTS = "mgmts"
    MGMT_ID = "mgmt_id"
    MGMT_NAME = "mgmt_name"
    MIN = "min"
    MINIP = "minIp"
    MIN_IP = "min_ip"
    MIN_ICMP_TYPE = "min_icmp_type"
    MIN_PORT = "min_port"
    MIN_PROTOCOL = "min_protocol"
    MIN_VALUE_SOURCE = "min_value_source"
    MODEL = "model"
    MODIFIED = "modified"
    MODIFIED_OBJECT_NAME = "modified_object_name"
    MODULE = "module"
    MODULE_AND_POLICY = "module_and_policy"
    MODULES_AND_POLICY = "modules_and_policy"
    MPLSINPUTLABEL = "mplsInputLabel"
    MPLSOUTPUTLABEL = "mplsOutputLabel"
    MULTI_ACCESS_REQUESTS = "multi_access_requests"
    MULTI_GROUP_CHANGE = "multi_group_change"
    MUSTCONTAIN = "mustContain"
    NAME = "name"
    NAT_INFO = "nat_info"
    NATLIST = "natList"
    NAT_RULE = "nat_rule"
    NAT_RULES = "nat_rules"
    NEGATE = "negate"
    NEGATED = "negated"
    NETMASK = "netmask"
    NETWORK = "network"
    NETWORK_ADDRESS = "network_address"
    NETWORK_COLLECTION = "network_collection"
    NETWORK_IP = "network_ip"
    NETWORK_ITEM = "network_item"
    NETWORK_ITEMS = "network_items"
    NETWORK_MASK = "network_mask"
    NETWORK_NAME = "network_name"
    NETWORK_OBJECT = "network_object"
    NETWORK_OBJECTS = "network_objects"
    NETWORKS = "networks"
    NETWORK_UID = "network_uid"
    NEW_REVISION = "new_revision"
    NEW_RULE = "new_rule"
    NEW_RULES = "new_rules"
    NEW_RULES_VIOLATED_TRAFFIC = "new_rules_violated_traffic"
    NEXT_HOP_IP = "next_hop_ip"
    NEXTHOPIP = "nextHopIp"
    NEXTDEVICES = "nextDevices"
    NOT_ALLOWED_SERVICE = "not_allowed_service"
    NOT_ALLOWED_SERVICES = "not_allowed_services"
    NOT_BLOCKED_SERVICE = "not_blocked_service"
    NOT_BLOCKED_SERVICES = "not_blocked_services"
    NOTES = "notes"
    NUMBER = "number"
    NUMBER_OF_RULES = "number_of_rules"
    OBJECT = "object"
    OBJECT_DETAILS = "object_details"
    OBJECT_NAME = "object_name"
    OBJECTNAMES = "objectNames"
    OBJECT_TYPE = "object_type"
    OBJECT_UID = "object_UID"
    OFFLINE = "offline"
    OFFLINE_DEVICE = "offline_device"
    OLD_REVISION = "old_revision"
    OLD_RULE = "old_rule"
    OLD_RULES = "old_rules"
    OLD_RULES_VIOLATED_TRAFFIC = "old_rules_violated_traffic"
    OPEN_TICKETS = "open_tickets"
    OPTIONS = "options"
    OPTION = "option"
    ORDER = "order"
    ORIG_DST_NETWORK = "orig_dst_network"
    ORIG_SRC_NETWORK = "orig_src_network"
    ORIG_SERVICE = "orig_service"
    ORIGIN = "origin"
    ORIGINAL_INSTANCE_ID = "original_instance_id"
    ORIGINALIPS = "originalIps"
    ORIGINALSERVICES = "originalServices"
    OUTGOING_INTERFACE_NAME = "outgoing_interface_name"
    OUTGOINGINTERFACE = "outgoingInterface"
    OUTGOINGINTERFACENAME = "outgoingInterfaceName"
    OUTGOINGVRF = "outgoingVrf"
    OUT_OF_OFFICE_FROM = "out_of_office_from"
    OUT_OF_OFFICE_UNTIL = "out_of_office_until"
    OWNER = "owner"
    OWNER_USER_ID = "owner_user_id"
    OWNER_DISPLAY_NAME = "owner_display_name"
    PATH = "path"
    PATH_CALC_RESULTS = "path_calc_results"
    PARTIAL_LIST = "partial_list"
    PARTICIPANT_USERNAME = "participant_username"
    PARTICIPANT_ID = "participant_id"
    PARTICIPANTS = "participants"
    PARTICIPATINGGATEWAYS = "participatingGateways"
    PAYLOAD = "payload"
    PEER = "peer"
    PERMISSIVENESS_LEVEL = "permissiveness_level"
    PHONE = "phone"
    POLICIES = "policies"
    POLICY = "policy"
    POLICYPACKAGE = "policyPackage"
    POLICYRULENUMBER = "policyRuleNumber"
    POLICY_ANALYSIS_QUERY_RESULT = "policy_analysis_query_result"
    POLICY_NAME = "policy_name"
    POLICY_CONTROL_NAME = "policy_control_name"
    POOL_MEMBER = "pool_member"
    POOL_MEMBERS = "pool_members"
    PORT = "port"
    PORT_FROM = "port_from"
    PERCENT_IMPLEMENTED = "percent_implemented"
    PERFORMED_BY = "performed_by"
    PRECEDENCE = "precedence"
    PREDEFINED_NAME = "predefined_name"
    PREDEFINED_SERVICE_ID = "predefined_service_id"
    PREDEFINED_SERVICE_NAME = "predefined_service_name"
    PREDEFINED_SERVICE_RANGE = "predefined_service_range"
    PREDEFINED_SERVICE_RANGES = "predefined_service_ranges"
    PRENAT_IFACE = "prenat_iface"
    POSTNAT_IFACE = "postnat_iface"
    PREFIX = "prefix"
    PRIMARY = "primary"
    PRIORITY = "priority"
    PRODUCTS = "products"
    PROTOCOL = "protocol"
    PUSH_STATUS = "push_status"
    QUERY_PARAMS = "query_params"
    RANGE_FIRST_IP = "range_first_ip"
    RANGE_LAST_IP = "range_last_ip"
    READY = "ready"
    READ_ONLY = "read_only"
    REASON = "reason"
    REASSIGN_TASK_COMMENT = "reassign_task_comment"
    RECIPIENT = "recipient"
    RECIPIENTS = "recipients"
    RECIPIENTS_ANY = "recipients_any"
    RECORD_SET = "record_set"
    REDONE = "redone"
    REDO_STEP_COMMENT = "redo_step_comment"
    REFERENCED = "referenced "
    REGION = "region"
    REMEDIATION = "remediation"
    REPORT = "report"
    REPORTS = "reports"
    RESULT = "result"
    REQUESTED_BY = "requested_by"
    REQUESTER = "requester"
    REQUESTER_ID = "requester_id"
    REVISION = "revision"
    REVISION_NUMBER = "revision_number"
    REVISION_ID = "revision_id"
    REVISIONID = "revisionId"
    REVISIONS = "revisions"
    RISK = "risk"
    RISK_ANALYSIS_RESULT = "risk_analysis_result"
    ROLE = "role"
    ROLES = "roles"
    ROUTE = "route"
    ROUTEDESTINATION = "routeDestination"
    ROUTES = "routes"
    RULE = "rule"
    RULES = "rules"
    RULE_COUNT = "rule_count"
    RULE_DESCRIPTION = "rule_description"
    RULE_ID = "rule_id"
    RULEIDENTIFIER = "ruleIdentifier"
    RULE_LOCATION = "rule_location"
    RULE_METADATA = "rule_metadata"
    RULE_NUMBER = "rule_number"
    RULE_ORDER = "rule_order"
    RULE_PLACMENT = "rule_placement"
    RULE_VIOLATED_TRAFFIC = "rule_violated_traffic"
    RULENUMBER = "ruleNumber"
    RULE_PROPERTIES = "rule_properties"
    RULE_PROPERTY = "rule_property"
    RULE_PROPERTIES_VIOLATIONS = "rule_properties_violations"
    RULE_PROPERTY_VIOLATION = "rule_property_violation"
    RULE_TEXT = "textual_rep"
    SATELLITEGATEWAYS = "satelliteGateways"
    SECURE_APP_APPLICATION = "secure_app_application"
    SEQNUMBER = "seqNumber"
    SCORE = "score"
    SCHEDULING = "scheduling"
    SECURITY_GROUP = "security_group"
    SECURITY_GROUPS = "security_groups"
    SECURITY_POLICY = "securityPolicy"
    SECURITY_POLICY_LIST = "SecurityPolicyList"
    SECURITY_POLICY_DEVICE_VIOLATIONS = "security_policy_device_violations"
    SECURITY_POLICY_EXCEPTION = "security_policy_exception"
    SECURITY_POLICY_EXCEPTION_LIST = "security_policy_exception_list"
    SECURITY_POLICY_VIOLATION = "security_policy_violation"
    SECURITY_POLICY_VIOLATIONS = "security_policy_violations"
    SECURITY_REQUIREMENT = "security_requirement"
    SECURITY_REQUIREMENTS = "security_requirements"
    SECURITY_RULE_COUNT = "security_rule_count"
    SECURITY_ZONE_MATRIX = "security_zone_matrix"
    SELECTED_OPTION = "selected_option"
    SELECTED_OPTIONS = "selected_options"
    SELECTION = "selection"
    SEND_EMAIL = "send_email"
    SERVER = "server"
    SERVERS = "servers"
    SERVER_DECOMMISSION_REQUEST = "server_decommission_request"
    SERVICE = "service"
    SERVICENEGATED = "serviceNegated"
    SERVICES = "services"
    SERVICE_COLLECTION = "service_collection"
    SERVICE_ITEM = "service_item"
    SERVICE_ITEMS = "service_items"
    SERVICE_NAME = "service_name"
    SERVICE_UID = "service_uid"
    SERVICENATMETHOD = "serviceNatMethod"
    SEVERITY = "severity"
    SHADOWED_RULE = "shadowed_rule"
    SHADOWED_RULES = "shadowed_rules"
    SHADOWED_RULES_CLEANUP = "shadowed_rules_cleanup"
    SHADOWED_STATUS = "shadowed_status"
    SHADOWING_RULES = "shadowing_rules"
    SHARED = "shared"
    SKIPPED = "skipped"
    SLA_OUTCOME = "sla_outcome"
    SLA_STATUS = "sla_status"
    SOURCE = "source"
    SOURCES = "sources"
    SOURCEIP = "sourceIp"
    SOURCE_DOMAIN = "source_domain"
    SOURCE_NETWORK_COLLECTION = "source_network_collection"
    SOURCES_IN_ZONE = "sources_in_zone"
    SOURCENETWORKS = "sourceNetworks"
    SOURCENEGATED = "sourceNegated"
    SOURCESERVICES = "sourceServices"
    SRC = "src"
    SRC_NETWORK = "src_network"
    SRC_NETWORKS = "src_networks"
    SRC_NETWORKS_NEGATED = "src_networks_negated"
    SRC_SERVICES_NEGATED = "src_services_negated"
    SRCNATMETHOD = "srcNatMethod"
    STATUS = "status"
    STEP = "step"
    STEP_NAME = "step_name"
    STEPS = "steps"
    SUBNET = "subnet"
    SUBNET_MASK = "subnet_mask"
    SUBJECT = "subject"
    SUGGESTIONS_PER_BINDING = "suggestions_per_binding"
    SUGGESTIONS_PER_DEVICE = "suggestions_per_device"
    TAG = "tag"
    TAG_SERVERS = "tag_servers"
    TAGS = "tags"
    TAGS_SERVERS = "tags_servers"
    TARGET = "target"
    TARGETS = "targets"
    TASK = "task"
    TASK_NAME = "task_name"
    TASKS = "tasks"
    TECH_OWNER = "tech_owner"
    TESTDEF = "testDef"
    TESTDEFUID = "testDefUid"
    TESTPARAMS = "testParams"
    TEST_PRODUCTS = "test_products"
    TESTUID = "testUid"
    TEXT = "text"
    TEXT_AREA = "text_area"
    TEXT_FIELD = "text_field"
    TICKET = "ticket"
    TICKETS = "tickets"
    TICKET_HISTORY_ACTIVITIES = "ticket_history_activities"
    TICKET_HISTORY_ACTIVITY = "ticket_history_activity"
    TICKET_ID = "ticket_id"
    TICKET_IDS = "ticket_ids"
    TICKETCR = "ticketCr"
    TIME = "time"
    TIMEOUT = "timeout"
    TOPOLOGY = "topology"
    TO = "to"
    TO_DOMAIN = "to_domain"
    TO_ZONE = "to_zone"
    TOTAL = "total"
    TRACK = "track"
    TRACK_LEVEL = "track_level"
    TRACK_INTERVAL = "track_interval"
    TRAFFIC_RANGE = "traffic_range"
    TRAFFIC_ALLOWED = "traffic_allowed"
    TRANSLATEDIPS = "translatedIps"
    TRANSLATEDSERVICES = "translatedServices"
    TRANSLATED_SERVICE = "translated_service"
    TRANSLATED_DST_NETWORK = "translated_dst_network"
    TRANSLATED_SRC_NETWORK = "translated_src_network"
    TYPE = "type"
    TYPE_ON_DEVICE = "type_on_device"
    UID = "uid"
    UNAUTHORIZED_OPENED_ACCESS = "unauthorized_opened_access"
    UNAUTHORIZED_CLOSED_ACCESS = "unauthorized_closed_access"
    UNIQUE_ACTIVE_IN_ITG = "unique_active_in_itg"
    URL = "url"
    USER = "user"
    USER_ID = "user_id"
    USERS = "users"
    USAGE_MODE = "usage_mode"
    VALUE = "value"
    VENDOR = "vendor"
    VENDOR_NAME = "vendor_name"
    VERIFIED = "verified"
    VERIFICATION_STATUS = "verification_status"
    VERIFIER_BINDING = "verifier_binding"
    VERIFIER_BINDINGS = "verifier_bindings"
    VERIFIER_RESULT = "verifier_result"
    VERIFIER_TARGET = "verifier_target"
    VERIFIER_TARGETS = "verifier_targets"
    VERIFIER_WARNING = "verifier_warning"
    VERSION_ID = "version_id"
    VIEWER = "viewer"
    VIEWERS = "viewers"
    VISIBILITY = "visibility"
    VIOLATED_TRAFFIC = "violated_traffic"
    VIOLATING_SERVICE = "violating_service"
    VIOLATING_SERVICES = "violating_services"
    VIOLATION = "violation"
    VIOLATIONS = "violations"
    VIOLATING_OBJECT = "violating_object"
    VIOLATING_RULE = "violating_rule"
    VIOLATING_RULES = "violating_rules"
    VIRTUAL_IP = "virtual_ip"
    VIRTUAL_ROUTER = "virtual_router"
    VM_INSTANCE = "vm_instance"
    VM_INSTANCES = "vm_instances"
    VPN = "vpn"
    VPNCONNECTION = "vpnConnection"
    WORKFLOW = "workflow"
    ZONE = "zone"
    ZONES = "zones"
    ZONE_ID = "zone_id"
    ZONEID = "zoneId"
    ZONE_ENTRY = "zone_entry"
    ZONE_ENTRIES = "zone_entries"
    ZONE_NAME = "zone_name"
    ZONE_NAME_IN_PARENT = "zone_name_in_parent"
    ZONE_TO_ZONE_SECURITY_REQUIREMENT = "zone_to_zone_security_requirement"


@enum.unique
class SeverityLevels(enum.Enum):
    CRITICAL = "CRITICAL"
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"

    @staticmethod
    def from_string(level_string):
        """
        Get enum from string
        :type level_string: str
        :return:
        """
        try:
            level = [level for level in (SeverityLevels.CRITICAL, SeverityLevels.HIGH,
                                         SeverityLevels.MEDIUM, SeverityLevels.LOW) if level.value == level_string][0]
        except IndexError:
            raise ValueError("String {} doesn't represent any of the severity levels".format(level_string))
        else:
            return level

    def _compare_with_enum(self, level):
        """
        Check if greater than other provided violation severity level
        :type level: SeverityLevels
        :return:
        """
        if self == SeverityLevels.CRITICAL and level == SeverityLevels.CRITICAL:
            return False
        elif self == SeverityLevels.HIGH and level in (SeverityLevels.HIGH, SeverityLevels.CRITICAL):
            return False
        elif self == SeverityLevels.MEDIUM and level in (SeverityLevels.CRITICAL,
                                                         SeverityLevels.HIGH, SeverityLevels.MEDIUM):
            return False
        if self == SeverityLevels.LOW:
            return False
        return True

    def __gt__(self, level):
        """
        Implementation of greater
        :param level:
        :return:
        """
        if isinstance(level, str):
            level = SeverityLevels.from_string(level)
            return self._compare_with_enum(level)
        elif isinstance(level, SeverityLevels):
            return self._compare_with_enum(level)
        else:
            raise TypeError("Wrong type to compare")



from pytos.securechange.xml_objects.restapi.step.access_request.accessrequest import Any_Access_Request_Device, \
    Named_Access_Request_Device, IP_Range_Access_Request_Target, IP_Access_Request_Target, Object_Access_Request_Target, \
    Internet_Access_Request_Target, Any_Access_Request_Target, DNS_Access_Request_Target, \
    LDAP_Entity_Access_Request_Target
from pytos.securechange.xml_objects.restapi.step.initialize import *

logger = logging.getLogger(XML_LOGGER_NAME)


class Step_Field_Server_Decommission(Step_Multi_Field_Base):
    FIELD_CONTENT_ATTRIBUTES = "server_decommission_request"

    def __init__(self, num_id, name, server_decommission_requests, read_only=None):
        if server_decommission_requests is None:
            server_decommission_requests = []
        self.server_decommission_requests = server_decommission_requests
        super().__init__(num_id, name, read_only)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.FIELD_TYPE_MULTI_SERVER_DECOMMISSION_REQUEST)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        read_only = get_xml_text_value(xml_node, Elements.READ_ONLY)
        server_decommission_requests = []
        for server_decommission_request_node in xml_node.findall(Elements.SERVER_DECOMMISSION_REQUEST):
            server_decommission_request = ServerDecommissionRequest.from_xml_node(server_decommission_request_node)
            server_decommission_requests.append(server_decommission_request)

        return cls(num_id, name, server_decommission_requests, read_only)

    def to_pretty_str(self):
        # TODO: Implement this function
        pass


class ServerDecommissionRequest(XML_Object_Base):
    def __init__(self, server_decommission_id, order, targets, servers, comment, domain,
                 designer_result=None, verifier_result=None,
                 impact_analysis_result=None, reado_only=None):
        self.id = server_decommission_id
        self.order = order
        self.targets = XML_List(Elements.TARGETS, targets)
        self.servers = XML_List(Elements.SERVERS, servers)
        self.comment = comment
        self.designer_results = designer_result
        self.verifier_result = verifier_result
        self.impact_analysis_result = impact_analysis_result
        if domain:
            self.domain = domain
        self.read_only = reado_only
        super().__init__(Elements.SERVER_DECOMMISSION_REQUEST)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        order = get_xml_text_value(xml_node, Elements.ORDER)
        server_decommission_id = get_xml_int_value(xml_node, Elements.ID)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        reado_only = get_xml_text_value(xml_node, Elements.READ_ONLY)
        device_type_to_class_dict = {TYPE_ANY: Any_Access_Request_Device, TYPE_OBJECT: Named_Access_Request_Device}
        targets = XML_List.from_xml_node_by_type_dict(xml_node, Elements.TARGETS, Elements.TARGET,
                                                      device_type_to_class_dict)

        server_type_to_class_dict = {TYPE_RANGE: IP_Range_Access_Request_Target, TYPE_IP: IP_Access_Request_Target,
                                     TYPE_DNS: DNS_Access_Request_Target, TYPE_OBJECT: Object_Access_Request_Target,
                                     TYPE_ANY: Any_Access_Request_Target, TYPE_INTERNET: Internet_Access_Request_Target,
                                     TYPE_LDAP_ENTITY: LDAP_Entity_Access_Request_Target}

        servers = XML_List.from_xml_node_by_type_dict(xml_node, Elements.SERVERS, Elements.SERVER,
                                                      server_type_to_class_dict)

        domain = get_xml_text_value(xml_node, Elements.DOMAIN)

        # TODO: Implement

        designer_result = None
        verifier_result = None
        impact_analysis_result = None

        return cls(server_decommission_id, order, targets, servers, comment, domain,
                   designer_result, verifier_result, impact_analysis_result, reado_only)

    def to_pretty_str(self):
        server_decommission_request_string = "Server Decommission {}:\n".format(self.order)
        server_decommission_request_string += "\tTargets: "
        for target in self.targets:
            server_decommission_request_string += target.to_pretty_str()
            server_decommission_request_string += "\n\tServers: "
        for server in self.servers:
            server_decommission_request_string += server.to_pretty_str()
        if self.comment is not None:
            server_decommission_request_string += "\n" + textwrap.fill(
                "\tComment: {}".format(unescape(self.comment)),
                initial_indent='', subsequent_indent='\t\t ')
        return server_decommission_request_string

    def __str__(self):
        return self.to_pretty_str()




from pytos.securechange.xml_objects.restapi.step.initialize import *
from pytos.securechange.xml_objects.restapi.step.step import AbsNetwork, AbsService, Binding

logger = logging.getLogger(XML_LOGGER_NAME)


class Analysis_Result(XML_Object_Base):
    IMPLEMENTED = "implemented"
    NOT_AVAILABLE = "not available"
    NOT_IMPLEMENTED = "not implemented"
    NOT_RUN = "not run"
    VERIFIED = "verified"

    def __init__(self, xml_tag, status):
        self.status = status
        super().__init__(xml_tag)

    def is_not_run(self):
        if self.status == Analysis_Result.NOT_RUN:
            return True
        else:
            return False

    def is_not_available(self):
        if self.status == Analysis_Result.NOT_AVAILABLE:
            return True
        else:
            return False

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        raise NotImplementedError(
                "from_xml_node must be implemented by derived classes.")


class Verifier_Result(Analysis_Result):
    """This class represents the Verifier DTO"""

    def __init__(self, status, result):
        self.status = status
        if result:
            self.result = result
        super().__init__(Elements.VERIFIER_RESULT, status)

    def is_implemented(self):
        return self.status == Verifier_Result.IMPLEMENTED

    def is_verified(self):
        return self.status == Verifier_Result.VERIFIED

    def is_not_implemented(self):
        return self.status == Verifier_Result.NOT_IMPLEMENTED

    def is_not_run(self):
        return self.status == Verifier_Result.NOT_RUN

    def is_not_available(self):
        return self.status == Verifier_Result.NOT_AVAILABLE

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        status = get_xml_text_value(xml_node, Elements.STATUS)
        result = get_xml_node(xml_node, Elements.RESULT, True)
        if result is not None:
            result = ReferenceURL.from_xml_node(result)
        return cls(status, result)


class AccessRequestVerifierResult(XML_Object_Base):
    """The class represents the expended verifier results """

    def __init__(self, verifier_targets):
        self.verifier_targets = verifier_targets
        super().__init__(Elements.ACCESS_REQUEST_VERIFIER_RESULT)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """

        verifier_targets = XML_List.from_xml_node_by_tags(xml_node, Elements.VERIFIER_TARGETS,
                                                          Elements.VERIFIER_TARGET, VerifierTarget)
        return cls(verifier_targets)


class VerifierTarget(XML_Object_Base):
    """This class represents the VerifierDeviceDTO"""

    def __init__(self, management_name, managementId, device_type, revision_number, administrator,
                 date, time, vendor, verification_status, verifier_bindings, verifier_warning):
        self.management_name = management_name
        self.managementId = managementId
        self.device_type = device_type
        self.revision_number = revision_number
        self.administrator = administrator
        self.date = date
        self.time = time
        self.vendor = vendor
        self.verification_status = verification_status
        self.verifier_bindings = verifier_bindings
        self.verifier_warning = verifier_warning
        super().__init__(Elements.VERIFIER_TARGET)

    def to_pretty_str(self):
        verifier_string = "Managment name: {}\n".format(self.management_name)
        verifier_string += "Revision Number: {}\n".format(self.revision_number)
        verifier_string += "Date: {}\n".format(self.date)
        verifier_string += "Administrator: {}\n".format(self.administrator)
        verifier_string += "Status: {}\n".format(self.verification_status)
        verifier_string += "\n\n"
        for verifier_binding in self.verifier_bindings:
            for element, value in verifier_binding.binding.__dict__.items():
                if element.startswith('_'):
                    continue
                verifier_string += "{}: {}\n\n".format(element.capitalize(), value)

            percent = verifier_binding.percent_implemented
            stat = "Implemented ({}%)\n\n".format(percent) if percent else "Not Implemented ({}%)\n\n".format(percent)
            verifier_string += "Status: {}".format(stat)

            verifier_string += "Violating Rules: "
            violating_rule_string = ''
            for violating_rule in verifier_binding.violating_rules:
                violating_rule_string += "\n\tSources: {}".format(violating_rule.src_networks.display_name)
                violating_rule_string += "\n\tDestinations: {}".format(violating_rule.dst_networks.display_name)
            if not violating_rule_string:
                msg = "The queried traffic is handled by the implicit cleanup rule."
                verifier_string += "---" if self.verification_status.lower() == "implemented" else msg
            else:
                verifier_string += violating_rule_string
            verifier_string += "\n\n"

            verifier_string += "Implementing Rules: "
            implementing_rule_string = ''
            for implementing_rule in verifier_binding.implementing_rules:
                implementing_rule_string += "\n\tSources: {}/{}".format(implementing_rule.src_networks.ip,
                                                                        implementing_rule.src_networks.subnet_mask)
                implementing_rule_string += "\n\tDestinations: {}/{}".format(implementing_rule.dst_networks.ip,
                                                                             implementing_rule.dst_networks.subnet_mask)
                implementing_rule_string += "\n\tServices: {}".format(implementing_rule.dst_service.name)
            if not implementing_rule_string:
                verifier_string += "--------"
            else:
                verifier_string += implementing_rule_string
            verifier_string += "\n\n-----------------------------------------------------------\n\n\n"
        verifier_string += "\n"
        return verifier_string

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        management_name = get_xml_text_value(xml_node, Elements.MANAGEMENT_NAME)
        management_id = get_xml_int_value(xml_node, Elements.MANAGEMENT_ID)
        device_type = get_xml_text_value(xml_node, Elements.DEVICE_TYPE)
        revision_number = get_xml_int_value(xml_node, Elements.REVISION_NUMBER)
        administrator = get_xml_text_value(xml_node, Elements.ADMINISTRATOR)
        date = get_xml_text_value(xml_node, Elements.DATE)
        time = get_xml_text_value(xml_node, Elements.TIME)
        vendor = get_xml_text_value(xml_node, Elements.VENDOR)
        verifier_warning = get_xml_text_value(xml_node, Elements.VERIFIER_WARNING)
        verification_status = get_xml_text_value(xml_node, Elements.VERIFICATION_STATUS)
        verifier_bindings = XML_List.from_xml_node_by_tags(xml_node, Elements.VERIFIER_BINDINGS,
                                                           Elements.VERIFIER_BINDING, VerifierBinding)
        return cls(management_name, management_id, device_type, revision_number, administrator,
                   date, time, vendor, verification_status, verifier_bindings, verifier_warning)


class VerifierBinding(XML_Object_Base):
    """This class represents the VerifierBindingDTO"""

    def __init__(self, handled_by_implicit_cleanup_rule, percent_implemented, implementing_rules,
                 implementation_percentage_threshold, verified, verifier_warning, violating_rules, binding):
        self.handled_by_implicit_cleanup_rule = handled_by_implicit_cleanup_rule
        self.percent_implemented = percent_implemented
        self.implementing_rules = implementing_rules
        self.implementation_percentage_threshold = implementation_percentage_threshold
        self.verified = verified
        self.verifier_warning = verifier_warning
        self.violating_rules = violating_rules
        self.binding = binding
        super().__init__(Elements.VERIFIER_BINDING)

    @classmethod
    def from_xml_node(cls, xml_node):
        handled_by_implicit_cleanup_rule = get_xml_text_value(xml_node, Elements.HANDLE_IMPLICIT_CLEANUP_RULE)
        percent_implemented = get_xml_text_value(xml_node, Elements.PERCENT_IMPLEMENTED)
        implementation_percentage_threshold = get_xml_text_value(xml_node, Elements.IMPLEMENTATION_PERCENTAGE_THRESHOLD)
        verified = get_xml_text_value(xml_node, Elements.VERIFIED)
        verifier_warning = get_xml_text_value(xml_node, Elements.VERIFIER_WARNING)
        violating_rules = XML_List.from_xml_node_by_tags(xml_node, Elements.VIOLATING_RULES,
                                                         Elements.VIOLATING_RULE, ViolatingRule)
        implementing_rules = XML_List.from_xml_node_by_tags(xml_node, Elements.IMPLEMENTING_RULES,
                                                            Elements.IMPLEMENTING_RULE, ImplementingRule)
        binding = Binding.from_xml_node(get_xml_node(xml_node, Elements.BINDING))
        return cls(handled_by_implicit_cleanup_rule, percent_implemented, implementing_rules,
                   implementation_percentage_threshold, verified, verifier_warning, violating_rules, binding)


class SlimRuleObject(XML_Object_Base):
    """This class represents the SlimRuleDTO"""

    def __init__(self, xml_node, target_type_tag):
        self.rule_number = get_xml_text_value(xml_node, Elements.RULENUMBER)
        self.dst_networks = SlimRuleObject.get_obj(AbsNetwork, xml_node, Elements.DESTNETWORKS)
        self.src_networks = SlimRuleObject.get_obj(AbsNetwork, xml_node, Elements.SOURCENETWORKS)
        self.src_service = SlimRuleObject.get_obj(AbsService, xml_node, Elements.SOURCESERVICES)
        self.dst_service = SlimRuleObject.get_obj(AbsService, xml_node, Elements.DESTINATIONSERVICES)
        self.action = get_xml_text_value(xml_node, Elements.ACTION)
        self.comment = get_xml_text_value(xml_node, Elements.COMMENT)
        self.name = get_xml_text_value(xml_node, Elements.NAME)
        super().__init__(target_type_tag)

    @staticmethod
    def get_obj(cls, xml_node, element):
        logger.debug("Findign the element '{}' in the XML".format(element))
        node = get_xml_node(xml_node, element, True)
        logger.debug("The node '{}' was found".format(node))
        if node is None:
            obj = Flat_XML_Object_Base(element)
        else:
            obj = cls.from_xml_node(node)
        return obj


class ImplementingRule(SlimRuleObject):
    """This class represents the implementing_rule element in the VerifierBindingDTO"""

    def __init__(self, xml_node):
        super().__init__(xml_node, Elements.IMPLEMENTING_RULE)

    @classmethod
    def from_xml_node(cls, xml_node):
        return cls(xml_node)


class ViolatingRule(SlimRuleObject):
    """This class represents the violating_rule element in the VerifierBindingDTO"""

    def __init__(self, xml_node):
        super().__init__(xml_node, Elements.VIOLATING_RULE)

    @classmethod
    def from_xml_node(cls, xml_node):
        return cls(xml_node)





import argparse
import codecs
import csv
import logging
import re
import string

from pytos.common.logging.definitions import COMMON_LOGGER_NAME

logger = logging.getLogger(COMMON_LOGGER_NAME)


def str_to_bool(bool_string):
    """
    Convert a string value to a corresponding boolean value
    :param bool_string: The string to convert.
    :type bool_string: str
    :rtype: bool
    """
    bool_string = bool_string.lower()
    if bool_string in ["yes", "y", "true", "t", "1", "enable"]:
        return True
    elif bool_string in ["no", "n", "false", "f", "0", "disable"]:
        return False
    else:
        raise ValueError("Cannot resolve '{}' to boolean value.".format(bool_string))


def read_multiline_str_from_stdin():
    """Read a multi-line string from STDIN"""
    input_string = ""
    while True:
        try:
            line = input()
            input_string += "{}\n".format(line)
        except EOFError:
            return input_string


def read_all_args():
    """Read all args from console
    :rtype: tuple"""
    parser = argparse.ArgumentParser()
    parser.add_argument("args", nargs="*")
    cli_args = parser.parse_args()
    return cli_args.args


def get_csv_parser(csv_file, encoding=None, delimiter=None, comment_char=None):
    """
    Read the contents of a file, ignoring any lines that begins with #, and returns a CSV parser for that file.
    :param csv_file: The path to the CSV file that will be parsed or buffer containing the CSV string.
    :type csv_file: str|io.BufferedIOBase
    :return: A CSV parser for the specified file.
    :rtype: _csv.reader
    """
    default_comment_char = "#"
    csv_strings_temp = []
    if comment_char is None:
        comment_char = default_comment_char
    try:
        if hasattr(csv_file, "read"):
            csv_file.seek(0)
            csv_strings = csv_file.read()
            if hasattr(csv_strings, "decode"):
                csv_strings = csv_strings.decode()
            csv_strings = csv_strings.split("\n")
        else:
            if encoding is not None:
                csv_file = codecs.open(csv_file, encoding=encoding)
            else:
                csv_file = open(csv_file)
            csv_strings = csv_file.readlines()
            csv_file.close()
    except (FileNotFoundError, PermissionError) as file_open_error:
        logger.error("Could not open file '%s', error was '%s'.", csv_file, file_open_error)
        raise file_open_error
    for line in csv_strings[:]:
        # Strip comment lines
        if not line.startswith(comment_char) and not line.isspace():
            if any(item.strip() for item in line.split(",")):
                csv_strings_temp.append(line)
    if delimiter is not None:
        csv_reader = csv.reader(csv_strings_temp, delimiter=delimiter)
    else:
        csv_reader = csv.reader(csv_strings_temp)
    return csv_reader


def strip_ansi_codes(ansi_str):
    """
    Strip ANSI codes from a string.
    :param ansi_str: The string from which to strip ANSI codes.
    :return: The string minus ANSI codes.
    :rtype: str
    """
    ansi_escape = re.compile(r'\x1b(?:\[\??|\(|\)|=)[0-9a-zA-Z;<=#]*')
    return ansi_escape.sub('', ansi_str)


def strip_non_printable(line):
    """
    Strip non-printable characters from a line.
    :param line: The line to clean up.
    :return: The stripped line.
    :rtype: str
    """
    line = strip_ansi_codes(line)
    line = ''.join(char for char in line if char in string.printable)
    return line




import logging
from datetime import datetime

from pytos.common.base_types import XML_Object_Base, XML_List, Comparable
from pytos.common.definitions import xml_tags
from pytos.common.functions import str_to_bool, XML_LOGGER_NAME
from pytos.common.functions.xml import get_xml_text_value, get_xml_int_value, get_xml_node

logger = logging.getLogger(XML_LOGGER_NAME)


class Device(XML_Object_Base, Comparable):
    def __init__(self, model, vendor, domain_id, domain_name, num_id, name, offline, topology=None, ip=None,
                 virtual_type=None, is_virtual=None):
        self.model = model
        self.vendor = vendor
        self.domain_name = domain_name
        self.domain_id = domain_id
        self.id = num_id
        self.name = name
        self.offline = offline
        self.topology = topology
        self.ip = ip
        self.virtual_type = virtual_type
        self._config = None
        self._children = []
        self._parent = None
        self._is_virtual = is_virtual
        super().__init__(xml_tags.Elements.DEVICE)

    def _key(self):
        return self.id, self.name

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        model = get_xml_text_value(xml_node, xml_tags.Elements.MODEL)
        num_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        domain_id = get_xml_int_value(xml_node, xml_tags.Elements.DOMAIN_ID)
        domain_name = get_xml_text_value(xml_node, xml_tags.Elements.DOMAIN_NAME)
        vendor = get_xml_text_value(xml_node, xml_tags.Elements.VENDOR)
        topology = get_xml_text_value(xml_node, xml_tags.Elements.TOPOLOGY)
        offline = get_xml_text_value(xml_node, xml_tags.Elements.OFFLINE)
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        return cls(model, vendor, domain_id, domain_name, num_id, name, offline, topology, ip)

    @classmethod
    def from_db_device(cls, db_device, domain_name):
        is_virtual = False
        if db_device.is_virtual_context or db_device.virtual_type == "virtual_context":
            is_virtual = True
        return cls(db_device.cp_type, db_device.management_type, db_device.customer_id, domain_name,
                   db_device.management_id, db_device.management_name, db_device.is_offline,
                   topology=db_device.has_topology, virtual_type=db_device.virtual_type,
                   is_virtual=is_virtual)

    def set_config(self, config):
        """

        :type config: str
        """
        self._config = config

    def get_config(self):
        return self._config

    def get_parent(self):
        return self._parent

    def get_parents_recursive(self):
        if self._parent is not None:
            parents = [self._parent]
            parents.extend(self._parent.get_parents_recursive())
            return parents
        else:
            return []

    def has_children(self):
        return bool(self._children)

    def get_children(self):
        return self._children

    def set_parent(self, parent):
        """

        :type parent: Device
        """
        self._parent = parent

    def set_children(self, children):
        """

        :type children: list[Device]|Devices_List
        """
        self._children = children

    def add_child(self, child):
        self._children.append(child)

    def __repr__(self):
        return "Device('{}','{}','{}','{}','{}','{}','{}','{}','{}',{})".format(self.model, self.vendor, self.domain_id,
                                                                                self.domain_name, self.id, self.name,
                                                                                self.offline, self.topology, self.ip,
                                                                                self._children)


class Devices_List(XML_List):
    """
    :type devices: list[Device]
    """

    def __init__(self, devices, total=None):
        self.count = len(devices)
        self.total = total
        super().__init__(xml_tags.Elements.DEVICES, devices)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        devices = []
        for device_node in xml_node.iter(tag=xml_tags.Elements.DEVICE):
            devices.append(Device.from_xml_node(device_node))
        total = get_xml_int_value(xml_node, xml_tags.Elements.TOTAL)
        return cls(devices, total)

    def extend(self, iterable):
        super().extend(iterable)
        self.count += iterable.count


class GenericDevice(XML_Object_Base):
    def __init__(self, device_id, device_name, domain_id):
        self.id = device_id
        self.name = device_name
        self.customer_id = domain_id
        super().__init__(xml_tags.Elements.DEVICE)

    @classmethod
    def from_xml_node(cls, xml_node):
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        device_name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        domain_id = get_xml_int_value(xml_node, xml_tags.Elements.CUSTOMER_ID)
        return cls(device_id, device_name, domain_id)


class GenericDevicesList(XML_List):
    def __init__(self, generic_devices):
        """Initialize the object from parameters."""
        super().__init__(xml_tags.Elements.GENERIC_DEVICES, generic_devices)

    @classmethod
    def from_xml_node(cls, xml_node):
        generic_devices = []
        for generic_device_node in xml_node.iter(tag=xml_tags.Elements.DEVICE):
            generic_devices.append(GenericDevice.from_xml_node(generic_device_node))
        return cls(generic_devices)


class Device_Revisions_List(XML_List):
    """
    :type revisions: list[Device_Revision]
    """

    def __init__(self, revisions):
        super().__init__(xml_tags.Elements.REVISIONS, revisions)
        self.sort()

    @classmethod
    def from_xml_node(cls, xml_node=None):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        revisions = []
        for revision_node in xml_node.iter(tag=xml_tags.Elements.REVISION):
            revisions.append(Device_Revision.from_xml_node(revision_node))
        return cls(revisions)

    def sort(self):
        self.set_contents(sorted(self.get_contents(), key=lambda revision: revision.revisionId))

    def get_latest_revision(self):
        if self.get_contents():
            self.sort()
            return self[-1]
        else:
            message = "Revision list is empty."
            logger.error(message)
            raise ValueError(message)


class Module_And_Policy(XML_Object_Base):
    def __init__(self, module, policy):
        self.module = module
        self.policy = policy
        super().__init__(xml_tags.Elements.MODULE_AND_POLICY)

    @classmethod
    def from_xml_node(cls, xml_node):
        module = get_xml_text_value(xml_node, xml_tags.Elements.MODULE)
        policy = get_xml_text_value(xml_node, xml_tags.Elements.POLICY)
        return cls(module, policy)


class Device_Revision(XML_Object_Base):
    AUTHORIZED_STATUSES = ("AUTOMATICALLY_AUTHORIZED", "MANUALLY_AUTHORIZED")
    UNAUTHORIZED_STATUSES = ("AUTOMATICALLY_UNAUTHORIZED", "MANUALLY_UNAUTHORIZED")
    OTHER_STATUSES = ("N_A", "CALCULATING", "ERROR", "PENDING")
    REVISION_DATE_FORMAT_STRING = "%Y-%m-%d"
    REVISION_TIME_FORMAT_STRING = "%H:%M:%S.%f"

    def __init__(self, action, num_id, admin, auditLog, authorizationStatus, revision_date, revision_time, gui_client,
                 revision_id, modules_and_policy, policy_package, ready):
        self.action = action
        self.admin = admin
        self.auditLog = auditLog
        self.authorizationStatus = authorizationStatus
        self.date = revision_date
        self.guiClient = gui_client
        self.id = num_id
        self.modules_and_policy = modules_and_policy
        self.policyPackage = policy_package
        self.revisionId = revision_id
        self.time = revision_time
        self.ready = ready
        super().__init__(xml_tags.Elements.REVISION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        action = get_xml_text_value(xml_node, xml_tags.Elements.ACTION)
        admin = get_xml_text_value(xml_node, xml_tags.Elements.ADMIN)
        auditLog = get_xml_text_value(xml_node, xml_tags.Elements.AUDITLOG)
        authorizationStatus = get_xml_text_value(xml_node, xml_tags.Elements.AUTHORIZATIONSTATUS)
        revision_date = get_xml_text_value(xml_node, xml_tags.Elements.DATE)
        gui_client = get_xml_text_value(xml_node, xml_tags.Elements.GUICLIENT)
        num_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        modules_and_policy_node = get_xml_node(xml_node, xml_tags.Elements.MODULES_AND_POLICY, optional=True)
        if modules_and_policy_node:
            modules_and_policy = []
            for node in modules_and_policy_node.iterfind('module_and_policy'):
                modules_and_policy.append(Module_And_Policy.from_xml_node(node))
        else:
            modules_and_policy = None
        policy_package = get_xml_text_value(xml_node, xml_tags.Elements.POLICYPACKAGE)
        revision_id = get_xml_int_value(xml_node, xml_tags.Elements.REVISIONID)
        revision_time = get_xml_text_value(xml_node, xml_tags.Elements.TIME)
        ready = get_xml_text_value(xml_node, xml_tags.Elements.READY)
        return cls(action, num_id, admin, auditLog, authorizationStatus, revision_date, revision_time, gui_client,
                   revision_id, modules_and_policy, policy_package, ready)

    def is_authorized(self):
        return self.authorizationStatus in Device_Revision.AUTHORIZED_STATUSES

    def is_ready(self):
        return str_to_bool(self.ready)

    def get_revision_date(self):
        try:
            revision_date = datetime.strptime(self.date, Device_Revision.REVISION_DATE_FORMAT_STRING)
            return revision_date
        except ValueError as valueerror:
            logger.error("Could not parse date string '%s' using format string '%s'", self.date,
                         Device_Revision.REVISION_DATE_FORMAT_STRING)
            raise valueerror

    def get_revision_datetime(self):
        datetime_str = "{} {}".format(self.date, self.time)
        datetime_frmt = "{} {}".format(Device_Revision.REVISION_DATE_FORMAT_STRING,
                                       Device_Revision.REVISION_TIME_FORMAT_STRING)
        try:
            return datetime.strptime(datetime_str, datetime_frmt)
        except ValueError as error:
            logger.error("Could not parse date time '{}' using format string '{}'".format(datetime_str, datetime_frmt))
            raise error



class RuleSearchDevice(XML_Object_Base):
    """Device object that return from rule_search API based on the rule documentation"""

    def __init__(self, device_id, revision_id, rule_count):
        self.device_id = device_id
        self.revision_id = revision_id
        self.rule_count = rule_count
        super().__init__(xml_tags.Elements.DEVICE)

    @classmethod
    def from_xml_node(cls, xml_node):
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        revision_id = get_xml_int_value(xml_node, xml_tags.Elements.REVISION_ID)
        rule_count = get_xml_int_value(xml_node, xml_tags.Elements.RULE_COUNT)
        return cls(device_id, revision_id, rule_count)


class RuleSearchDeviceList(XML_List):
    """List of effected devices based on the rule search API"""

    def __init__(self, devices):
        super().__init__(xml_tags.Elements.DEVICES, devices)

    @classmethod
    def from_xml_node(cls, xml_node):
        """Return only devices that the rule_count is greater than 0"""
        devices = []
        for device_node in xml_node.iter(tag=xml_tags.Elements.DEVICE):
            device_obj = RuleSearchDevice.from_xml_node(device_node)
            if int(device_obj.rule_count):
                devices.append(device_obj)
        return cls(devices)


class InternetReferralObject(XML_Object_Base):
    def __init__(self, device_id, object_name):
        super().__init__(xml_tags.Elements.INTERNET_REFERRAL_OBJECT)
        self.device_id = device_id
        self.object_name = object_name

    @classmethod
    def from_xml_node(cls, xml_node):
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        object_name = get_xml_text_value(xml_node, xml_tags.Elements.OBJECT_NAME)
        return cls(device_id, object_name)




import logging

from pytos.common.base_types import XML_Object_Base, XML_List
from pytos.common.definitions import xml_tags
from pytos.common.functions import get_xml_text_value
from pytos.common.logging.definitions import XML_LOGGER_NAME

logger = logging.getLogger(XML_LOGGER_NAME)


class Route(XML_Object_Base):
    def __init__(self, network_ip, network_mask, next_hop_ip, interface=None):
        self.network_ip = network_ip
        self.network_mask = network_mask
        self.next_hop_ip = next_hop_ip
        self.interface = interface
        super().__init__(xml_tags.Elements.ROUTE)

    @classmethod
    def from_xml_node(cls, xml_node):
        network_ip = get_xml_text_value(xml_node, xml_tags.Elements.NETWORK_IP)
        network_mask = get_xml_text_value(xml_node, xml_tags.Elements.NETWORK_MASK)
        next_hop_ip = get_xml_text_value(xml_node, xml_tags.Elements.NEXT_HOP_IP)
        interface = get_xml_text_value(xml_node, xml_tags.Elements.INTERFACE)
        return cls(network_ip, network_mask, next_hop_ip, interface)

    def __str__(self):
        return "Route({},{},{},{})".format(self.network_ip, self.network_mask, self.next_hop_ip, self.interface)


class RoutesList(XML_List):
    """
    :type routes: list[Route]
    """

    def __init__(self, routes):
        self.routes = routes
        super().__init__(xml_tags.Elements.ROUTES, routes)

    @classmethod
    def from_xml_node(cls, xml_node):
        routes = []
        for route_node in xml_node.iter(tag=xml_tags.Elements.ROUTE):
            routes.append(Route.from_xml_node(route_node))
        return cls(routes)



import codecs
import logging
import socket
import xml.etree.ElementTree as ET
import xml.sax.saxutils
from abc import ABCMeta, abstractmethod
from netaddr import IPSet

from pytos.common.definitions import xml_tags
from pytos.common.functions.xml import get_xml_node

SET_PARENT_NODE = "set_parent_node"
ID = "id"
PARENT_NODE = "_parent_node"

from pytos.common.logging.definitions import XML_LOGGER_NAME
logger = logging.getLogger(XML_LOGGER_NAME)


def _get_tab_string(element_level):
    tab_string = " " * XML_Base.SPACER_SIZE * element_level
    return tab_string


class Comparable:
    @abstractmethod
    def _key(self):
        raise NotImplementedError

    def __hash__(self):
        return hash(self._key())

    def __eq__(self, other):
        equals = True
        if not type(self) == type(other):
            raise TypeError("Uncomparable types '{}' and '{}'.".format(type(self), type(other)))
        elif self is other:
            return True
        else:
            for self_item, other_item in zip(self._key(), other._key()):
                if self_item != other_item:
                    equals = False
                    break
            if equals and super().__eq__(None) != NotImplemented:
                equals = super().__eq__(other)
            return equals

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            for self_item, other_item in zip(self._key(), other._key()):
                if self_item != other_item:
                    return self_item < other_item
            return False
        else:
            raise TypeError("Unorderable types: {} < {}".format(type(self), type(other)))


class XML_Base:
    SPACER_SIZE = 2

    def __init__(self, xml_tag, attribs=None):
        """
        :param xml_tag: The XML tag for the XML object.
        :type xml_tag: str
        :param attribs: The XML attributes of the XML_List.
        :type attribs: dict
        """
        if isinstance(xml_tag, str) or isinstance(xml_tag, int):
            self._xml_tag = xml_tag
        else:
            raise ValueError("xml_tag must be either a string or integer, xml_tag is '{}' of type '{}'".format(xml_tag,
                                                                                                               type(
                                                                                                                       xml_tag)))

        try:
            if self._attribs:
                try:
                    self._attribs.update(attribs)
                except TypeError:
                    self._attribs = {}
        except AttributeError:  # If set_attrib is called before __init__
            if attribs:
                self._attribs = attribs
            else:
                self._attribs = {}

    def get_xml_tag(self):
        """Get the XML tag for the current element."""
        return self._xml_tag

    def get_attribs(self):
        return self._attribs

    def get_attrib(self, attrib_name):
        if attrib_name in self._attribs:
            return self._attribs[attrib_name]

    def set_attrib(self, attrib_name, attrib_value):
        try:
            self._attribs[attrib_name] = attrib_value
        except AttributeError:
            self._attribs = {attrib_name: attrib_value}

    def to_xml_string(self, element_level=None):
        raise NotImplementedError("to_xml_string must be implemented by derived classes.")

    def to_xml_doc(self):
        raise NotImplementedError("to_xml_doc must be implemented by derived classes.")

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        raise NotImplementedError("from_xml_node must be implemented by derived classes.")

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Initialize the object from a XML string.
        :param xml_string: The string from which all necessary parameters will be parsed.
        :type xml_string: str
        """

        xml_node = ET.fromstring(xml_string)
        return cls.from_xml_node(xml_node)


class XML_List(XML_Base):
    """This class is used for when an XML element with one tag contains multiple other XML nodes with different tags."""

    def __init__(self, xml_tag, list_data=None, _attribs=None):
        """
        :param xml_tag: The XML tag for the XML_List.
        :type xml_tag: str
        :param list_data: The contents of the XML_List.
        :type list_data: list
        :param _attribs: The XML attributes of the XML_List.
        :type _attribs: dict
        """
        self._list_data = []
        if list_data:
            self._list_data.extend(list_data)
        super().__init__(xml_tag, _attribs)

    @classmethod
    def from_xml_node_by_tags(cls, xml_node, list_element_name, child_element_name, child_class_type, optional=False):
        xml_list_base_node = get_xml_node(xml_node, list_element_name, optional)
        list_data = []
        if xml_list_base_node:
            for child_node in xml_list_base_node.iterfind(child_element_name):
                list_data.append(child_class_type.from_xml_node(child_node))
        return cls(list_element_name, list_data)

    @classmethod
    def from_xml_node_by_type_dict(cls, xml_node, list_element_name, child_element_name, type_to_class_dict,
                                   optional=False, default_class=None):
        xml_list_base_node = get_xml_node(xml_node, list_element_name, optional)
        list_data = []
        if xml_list_base_node:
            for child_node in xml_list_base_node.iter(tag=child_element_name):
                try:
                    type_attribute = child_node.attrib[xml_tags.Attributes.XSI_NAMESPACE_TYPE]
                except KeyError:
                    try:
                        type_attribute = child_node.attrib[xml_tags.TYPE_ATTRIB]
                    except KeyError:
                        message = "could not find type attribute in {},existing attributes: {}".format(child_node,
                                                                                                       child_node.attrib)
                        logger.error(message)
                        raise KeyError(message)
                try:
                    list_data.append(type_to_class_dict[type_attribute].from_xml_node(child_node))
                except KeyError:
                    if default_class:
                        list_data.append(default_class.from_xml_node(child_node))
                    else:
                        logger.critical("Unknown element '{}'. Skipping it.".format(type_attribute))

        return cls(list_element_name, list_data)

    @classmethod
    def from_xml_node_by_tag_dict(cls, xml_node, list_element_name, tag_to_class_dict, default_class=None):
        list_data = []
        for child_node in xml_node:
            try:
                list_data.append(tag_to_class_dict[child_node.tag].from_xml_node(child_node))
            except KeyError as error:
                if default_class:
                    list_data.append(default_class.from_xml_node(child_node))
                else:
                    raise error
        return cls(list_element_name, list_data)

    def to_xml_string(self, element_level=None):
        if self.get_xml_tag():
            if element_level is not None:
                element_level += 1
            else:
                element_level = 0
            logger.debug("Element level is '%s'.", element_level)
            tab_string = _get_tab_string(element_level)
            if self._list_data:
                xml_string = "{tab_string}<{xml_tag}>".format(xml_tag=self.get_xml_tag(), tab_string=tab_string)
                non_xml_items = []
                for item in self._list_data:
                    try:
                        xml_string += "\n{}".format(item.to_xml_string(element_level))
                    except AttributeError:
                        non_xml_items.append(xml.sax.saxutils.escape(str(item)))
                if non_xml_items:
                    xml_string += "\n{}".format(",".join(non_xml_items))
                xml_string += "\n{tab_string}</{xml_tag}>".format(xml_tag=self.get_xml_tag(), tab_string=tab_string)
            else:
                xml_string = "\n{tab_string}<{xml_tag}/>".format(xml_tag=self.get_xml_tag(), tab_string=tab_string)
            return xml_string
        else:
            raise ValueError("{} must have a _xml_tag attribute in order to print in XML form.")

    def to_xml_doc(self):
        """
        Returns a Element Tree representation of the XML Object Base.
        :return: An XML representation of the object in xml.etree.Element form.
        :rtype: xml.etree.ElementTree.Element
        """
        return ET.fromstring(self.to_xml_string())

    def get_attribs(self):
        return self._attribs

    def append(self, item):
        return self._list_data.append(item)

    def remove(self, value):
        return self._list_data.remove(value)

    def index(self, index):
        return self._list_data.index(index)

    def insert(self, index, value):
        return self._list_data.insert(index, value)

    def extend(self, iterable):
        return self._list_data.extend(iterable)

    def pop(self, index=0):
        return self._list_data.pop(index)

    def __next__(self):
        return next(self._list_data)

    def __iter__(self):
        if self._list_data is not None:
            return self._list_data.__iter__()

    def __len__(self):
        return self._list_data.__len__()

    def __getitem__(self, index):
        return self._list_data.__getitem__(index)

    def __delitem__(self, index):
        return self._list_data.__delitem__(index)

    def __setitem__(self, index, value):
        return self._list_data.__setitem__(index, value)

    def __str__(self):
        return self.__repr__()

    def __bool__(self):
        return bool(self._list_data)

    def set_contents(self, contents):
        self._list_data = contents

    def get_contents(self):
        return self._list_data

    def __repr__(self):
        return "XML_List('{}', {})".format(self._xml_tag, [repr(item) for item in self])


class XML_Object_Base(XML_Base):
    """
    This class is used as the basis for all other REST objects, and is used to represent
    a Python class as an XML string.
    Each derived class has a self._xml_tag object that is used as the enclosing XML tag for that object.
    Each member represents another tag: For example, self.name would get printed <name></name>
    In the event that a member name conflicts with a Python reserved keyword, an underscore
    is appended to the member variable name,
    even though when the object is converted to an XML the underscore is not displayed.
    """

    def __init__(self, xml_tag, _attribs=None):
        super().__init__(xml_tag, _attribs)
        self._parent_node = None
        self._set_parent_node_for_child_nodes()

    def get_parent_node(self):
        return self._parent_node

    def set_parent_node(self, node):
        self._parent_node = node

    def sanitize_ids(self):
        """
        Deletes all instance attributes with the name "id" in current instance recursively.
        """
        if hasattr(self, ID):
            setattr(self, ID, None)
        for member in self.__dict__:
            if member != PARENT_NODE and isinstance(self.__dict__[member], XML_Base):
                try:
                    self.__dict__[member].sanitize_ids()
                except AttributeError:
                    pass

    def _set_parent_node_for_child_nodes(self):
        for key in self.__dict__:
            if isinstance(self.__dict__[key], XML_Object_Base):
                self.__dict__[key].set_parent_node(self)
            elif isinstance(self.__dict__[key], XML_List) or isinstance(self.__dict__[key], list):
                for subitem in self.__dict__[key]:
                    if hasattr(subitem, SET_PARENT_NODE):
                        subitem.set_parent_node(self)

    def get_nth_parent_node(self, parent_level):
        if not isinstance(parent_level, int) or parent_level < 1:
            raise ValueError("Parent level must be an integer with a value of at least 1.")
        item_node = self
        for _ in range(parent_level):
            try:
                item_node = item_node.get_parent_node()
            except AttributeError:
                raise ValueError("Object {} does not have a parent node.".format(item_node))
        return item_node

    def to_xml_string(self, element_level=None):
        """
        Returns an XML string representation of the XML Object Base.
        :return: An XML representation of the object in string form.
        :rtype: str
        """
        if self.get_xml_tag():
            if element_level is not None:
                element_level += 1
            else:
                element_level = 0
            logger.debug("Element level is '%s'.", element_level)
            item_tab_string = _get_tab_string(element_level + 1)
            attrib_string = ""
            for key, value in self.get_attribs().items():
                attrib_string += ' {}="{}"'.format(key, value)
            xml_string = "{tab_string}<{xml_tag}{attrib_string}>".format(xml_tag=self._xml_tag,
                                                                         attrib_string=attrib_string,
                                                                         tab_string=_get_tab_string(element_level))
            for key in sorted(self.__dict__):
                # Skip the self._xml_tag member and self._attribs of each class ( and any other member beginning with _.
                if key.startswith("_"):
                    continue
                # If the element contains data, handle normally.
                logger.debug("Handling item '%s'.", key)
                # Handle a case where the member itself does not have an XML tag, but it contains XML elements.
                if isinstance(self.__dict__[key], list):
                    logger.debug("Handling list '%s' in XML form.", self.__dict__[key])
                    non_xml_items = []
                    for item in self.__dict__[key]:
                        if item is not None:
                            logger.debug("Handling list item '%s' in XML form.", item)
                            try:
                                xml_string += "\n{}".format(item.to_xml_string(element_level))
                            except AttributeError:
                                non_xml_items.append(xml.sax.saxutils.escape(str(item)))
                    if non_xml_items:
                        xml_string += "\n<{tag}>{non_xml_items}</{tag}>".format(tag=key,
                                                                                non_xml_items=",".join(non_xml_items))
                else:
                    try:
                        item_xml_string = self.__dict__[key].to_xml_string(element_level)
                        xml_string += "\n{item_xml_string}".format(item_xml_string=item_xml_string)
                        logger.debug("Handling XML_Base '%s'.", key)
                    except AttributeError:
                        # Handle the case where the member name conflicts with a reserved Python keyword,
                        # so an underscore (_) is appended to member name to avoid the conflict.
                        logger.debug("Handling other object type '%s'", key)
                        key_name = key
                        if key_name.endswith("_"):
                            key_name = key_name[:-1]
                        if self.__dict__[key] is not None:
                            item_xml_string = xml.sax.saxutils.escape(str(self.__dict__[key]))
                            xml_string += "\n{tab_string}<{xml_tag}>{item_xml_string}</{xml_tag}>".format(
                                    xml_tag=key_name, item_xml_string=item_xml_string, tab_string=item_tab_string)
                        else:
                            xml_string += "\n{tab_string}<{xml_tag}/>".format(xml_tag=key_name,
                                                                              tab_string=item_tab_string)
            xml_string += "\n{tab_string}</{xml_tag}>".format(xml_tag=self.get_xml_tag(),
                                                              tab_string=_get_tab_string(element_level))
            return xml_string
        else:
            raise ValueError("{} must have a _xml_tag attribute in order to output to XML form.")

    def to_xml_doc(self):
        """
        Returns a Element Tree representation of the XML Object Base.
        :return: An XML representation of the object in xml.etree.Element form.
        :rtype: xml.etree.ElementTree.Element
        """
        return ET.fromstring(self.to_xml_string())

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        raise NotImplementedError("from_xml_node must be implemented by derived classes.")

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Initialize the object from a XML string.
        :param xml_string: The string from which all necessary parameters will be parsed.
        :type xml_string: str
        """

        xml_node = ET.fromstring(xml_string)
        return cls.from_xml_node(xml_node)

    @classmethod
    def from_file(cls, xml_path):
        """
        Read an XML from a file and create an object from it.
        :param xml_path: The path to the file containing the XML.
        :type xml_path: str
        :return: The created object.
        :raise FileNotFoundError: If the template file was not found.
        """
        try:
            with codecs.open(xml_path, encoding='utf-8') as xml_file:
                xml_string = xml_file.read()
                xml_node = ET.fromstring(xml_string)
        except FileNotFoundError:
            message = "The file {} does not exist.".format(xml_path)
            logger.error(message)
            raise FileNotFoundError(message)
        else:
            return cls.from_xml_node(xml_node)


class Flat_XML_Object_Base(XML_Base):
    """
    This is the class for XML contains only single parameter and does not contain any tags inside
    """

    def __init__(self, xml_tag, _attribs=None, content=None):
        super().__init__(xml_tag, _attribs)
        self.content = content
        self._parent_node = None

    def get_parent_node(self):
        return self._parent_node

    def set_parent_node(self, node):
        self._parent_node = node

    def to_xml_string(self, element_level=None):
        """
        Returns an XML string representation of the Flat XML Object Base.
        :return: An XML representation of the object in string form.
        :rtype: str
        """
        if self.get_xml_tag():
            if element_level is not None:
                element_level += 1
            else:
                element_level = 0
            logger.debug("Element level is '%s'.", element_level)
            item_tab_string = _get_tab_string(element_level + 1)
            attrib_string = ""
            for key, value in self.get_attribs().items():
                attrib_string += ' {}="{}"'.format(key, value)
            xml_string = "{tab_string}<{xml_tag}{attrib_string}".format(tab_string=item_tab_string,
                                                                        xml_tag=self._xml_tag,
                                                                        attrib_string=attrib_string)
            if self.content:
                content = xml.sax.saxutils.escape(str(self.content))
                xml_string += ">{content}</{xml_tag}>".format(content=content, xml_tag=self._xml_tag)
            else:
                xml_string += "/>"
            return xml_string
        else:
            raise ValueError("{} must have a _xml_tag attribute in order to print in XML form.")

    def to_xml_doc(self):
        """
        Returns a Element Tree representation of the XML Object Base.
        :return: An XML representation of the object in xml.etree.Element form.
        :rtype: xml.etree.ElementTree.Element
        """
        return ET.fromstring(self.to_xml_string())


class XSI_Object(XML_Object_Base):
    def __init__(self, xml_tag, xsi_type):
        super().__init__(xml_tag)
        self.set_attrib(xml_tags.NAMESPACE_FIELD_ATTRIB_CONTENT, xml_tags.XSI_NAMESPACE_URL)
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xsi_type)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        raise NotImplementedError("from_xml_node must be implemented by derived classes.")


class ReferenceURL(XML_Object_Base):
    def __init__(self, url, tag_name):
        super().__init__(tag_name)
        self._url = url
        self.set_attrib(xml_tags.NAMESPACE_FIELD_ATTRIB_CONTENT, xml_tags.XSI_NAMESPACE_URL)
        self.set_attrib(xml_tags.Attributes.HREF, url)

    @classmethod
    def from_xml_node(cls, xml_node):
        url = xml_node.attrib[xml_tags.Attributes.HREF]
        return cls(url, xml_node.tag)

    @staticmethod
    def reference_from_xml_string(xml_string, class_ref):
        return class_ref.from_xml_string(xml_string)


class Service_Type:
    def __contains__(self, item):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    def __lt__(self, other):
        raise NotImplementedError

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def get_valid_port(port):
        if isinstance(port, str):
            try:
                port = int(socket.getservbyname(port.lower()))
            except OSError:
                raise ValueError("Service for port '{}' not found.".format(port.lower()))
        elif isinstance(port, int):
            if not 0 <= port <= 65535:
                raise ValueError("Port must be between 0 and 65535.")
        else:
            raise ValueError("Invalid port '{}'.".format(port))
        return port

    @staticmethod
    def get_valid_protocol(ip_protocol):
        if isinstance(ip_protocol, str):
            try:
                ip_protocol = socket.getprotobyname(ip_protocol.lower())
            except OSError:
                raise ValueError("Protocol '{}' not found.".format(ip_protocol.lower()))
        elif isinstance(ip_protocol, int):
            if not 0 <= ip_protocol <= 255:
                raise ValueError("Protocol must be between 0 and 255.")
        else:
            raise ValueError("Invalid IP protocol '{}'.".format(ip_protocol))
        return ip_protocol


class Single_Service_Type(Service_Type):
    def __init__(self, ip_protocol, port=0):
        self.ip_protocol = self.get_valid_protocol(ip_protocol)
        self.port = self.get_valid_port(port)

    def __eq__(self, other):
        """
        :param other:
        :return:
        """
        try:
            return self.ip_protocol == other.ip_protocol and self.port == other.port
        except AttributeError:
            return False

    def __contains__(self, item):
        return self == item

    def __hash__(self):
        return hash("{}/{}".format(self.ip_protocol, self.port))

    def __lt__(self, other):

        if type(other) == Any_Service_Type:
            return True
        elif hasattr(other, "_members"):
            return False
        elif hasattr(other, "start_port"):
            return self.ip_protocol < other.ip_protocol or (
                self.ip_protocol == other.ip_protocol and self.port < other.end_port)
        else:
            try:
                return self.ip_protocol < other.ip_protocol or (
                    self.ip_protocol == other.ip_protocol and self.port < other.port)
            except AttributeError:
                raise TypeError("Unorderable types: {} < {}".format(type(self), type(other)))

    def __repr__(self):
        return "Single_Service_Type({},{})".format(self.ip_protocol, self.port)


class Any_Service_Type(Service_Type):
    def __eq__(self, other):
        return type(other) == Any_Service_Type

    def __contains__(self, item):
        return True

    def __hash__(self):
        return hash("ANY")

    def __repr__(self):
        return "Any_Service_Type()"

    def __lt__(self, other):
        return False


class Range_Service_Type(Service_Type):
    def __init__(self, ip_protocol, start_port, end_port):
        self.ip_protocol = self.get_valid_protocol(ip_protocol)
        self.start_port = self.get_valid_port(start_port)
        self.end_port = self.get_valid_port(end_port)

    def __eq__(self, other):
        """
        :param other:
        :return:
        """
        try:
            return self.ip_protocol == other.ip_protocol and self.start_port == other.start_port and self.end_port ==\
                                                                                                     other.end_port
        except AttributeError:
            return False

    def __repr__(self):
        return "Range_Service_Type({},{},{})".format(self.ip_protocol, self.start_port, self.end_port)

    def __contains__(self, item):
        try:
            if hasattr(item, "_members"):
                for member in item.members:
                    if member in self:
                        continue
                    else:
                        return False
                return True
            elif self.ip_protocol == item.ip_protocol:
                if hasattr(item, "start_port"):
                    if self.start_port <= item.start_port and self.end_port >= item.end_port:
                        return True
                elif item.port in range(self.start_port, self.end_port + 1):
                    return True
                else:
                    return False

        except AttributeError:
            return False

    def __lt__(self, other):
        if type(other) == Any_Service_Type:
            return True

        elif hasattr(other, "_members"):
            return False
        elif hasattr(other, "start_port"):
            return self.ip_protocol < other.ip_protocol or (
                self.ip_protocol == other.ip_protocol and self.end_port < other.end_port) or (
                       self.ip_protocol == other.ip_protocol and self.end_port == other.end_port and self.start_port
                       < other.start_port)
        else:
            try:
                return self.ip_protocol < other.ip_protocol or (
                    self.ip_protocol == other.ip_protocol and self.end_port < other.port)
            except AttributeError:
                raise TypeError("Unorderable types: {} < {}".format(type(self), type(other)))

    def __hash__(self):
        return hash("{}/{}-{}".format(self.ip_protocol, self.start_port, self.end_port))


class Group_Service_Type(Service_Type):
    def __init__(self, members=None):
        if members is None:
            self._members = []
        else:
            self._members = members

    def __iter__(self):
        return iter(self._members)

    def __contains__(self, item):
        try:
            if hasattr(item, "_members"):
                for item_member in item:
                    if item_member not in self:
                        return False
                return True
            else:
                for member in self:
                    if item in member or item == member:
                        return True
                return False
        except AttributeError:
            return False

    def append(self, service):
        return self._members.append(service)

    def __lt__(self, other):
        if type(other) == Any_Service_Type:
            return True
        elif hasattr(other, '_members'):
            return False
        else:
            return True

    def __repr__(self):
        return "Group_Service_Type([{}])".format(",".join((repr(member) for member in self)))

    def __len__(self):
        return len(self._members)

    def __eq__(self, other):
        raise NotImplementedError


class Service_Set:
    def __init__(self, services=None):
        self._services = set()
        if services is not None:
            for service in services:
                if hasattr(service, "_members"):
                    self._services.update(service)
                else:
                    self._services.add(service)

    def add(self, item):
        self._services.add(item)

    def issubset(self, other_set):
        for service in self:
            if service not in other_set:
                return False
        return True

    def copy(self):
        return self.__class__(self._services)

    def __repr__(self):
        return "Service_Set([{}])".format(",".join((repr(service) for service in self)))

    def __iter__(self):
        return iter(self._services)

    def __contains__(self, other_service):
        if hasattr(other_service, "_members"):
            for member in other_service:
                if member not in self:
                    return False
            return True
        elif hasattr(other_service, '_services'):
            for service in other_service:
                if service not in self:
                    return False
            return True
        else:
            for service in self:
                if other_service in service or other_service == service:
                    return True
            return False

    def __len__(self):
        return len(self._services)


class Singleton(type):
    """
    Singleton Metaclass.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SubclassRegistry(type):
    """
    Metaclass that registers subclasses.
    """

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if not hasattr(cls, 'registry'):
            cls.registry = set()
        cls.registry.add(cls)
        cls.registry -= set(bases)  # Remove base classes

    # Metamethods, called on class objects:
    def __iter__(cls):
        return iter(cls.registry)


class SubclassWithIdentifierRegistry(type):
    """
    Metaclass that registers subclasses based on a class identifier.
    """
    CLASS_IDENTIFIER_VAR = "class_identifier"
    REGISTRY = 'registry'

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if not hasattr(cls, cls.REGISTRY):
            cls.registry = {}
        if hasattr(cls, cls.CLASS_IDENTIFIER_VAR):
            cls.registry[cls.class_identifier] = cls

    def __getitem__(cls, item):
        return cls.registry[item]


class IPNetworkMixin(metaclass=ABCMeta):
    def __init__(self):
        self._ip_network_cache = None

    @abstractmethod
    def _get_ip_network(self):
        """

        :rtype: netaddr.IPNetwork
        """
        pass

    def get_ip_network(self):
        """

        :rtype: netaddr.IPNetwork
        """
        if self._ip_network_cache is None:
            self._ip_network_cache = self._get_ip_network()
        return self._ip_network_cache

    def get_ip_set(self):
        """
        This returns an IPset representing the object.
        :rtype: IPSet
        """
        return IPSet(self.get_ip_network())





import logging

import netaddr

from pytos.common.base_types import XML_List, XML_Object_Base, Comparable, IPNetworkMixin
from pytos.common.definitions.xml_tags import Elements, Attributes
from pytos.common.functions import get_xml_int_value, get_xml_text_value, get_xml_node
from pytos.common.logging.definitions import XML_LOGGER_NAME

logger = logging.getLogger(XML_LOGGER_NAME)


class Zone_List(XML_List):
    def __init__(self, zones):
        """
        :type zones: list[Zone]
        """
        self.zones = zones
        self.count = len(self.zones)
        self.total = self.count
        super().__init__(Elements.ZONES, zones)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        zones = []
        for zone_node in xml_node.iter(Elements.ZONE):
            zones.append(Zone.from_xml_node(zone_node))
        return cls(zones)


class Zone_Domain(XML_Object_Base):
    def __init__(self, domain_id=1, name="Default"):
        self.id = domain_id
        self.name = name
        super().__init__(Elements.DOMAIN)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        domain_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        return cls(domain_id, name)


class Zone(XML_Object_Base, Comparable):
    def __init__(self, zone_id, name, comment, zone_domain=None):
        """
        :type zone_domain: Zone_Domain|None
        :type comment: str|None
        :type name: str
        :type zone_id: int|None
        """
        self.id = zone_id
        self.name = name
        self.comment = comment
        if zone_domain is None:
            zone_domain = Zone_Domain()
        self.domain = zone_domain
        super().__init__(Elements.ZONE)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        zone_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        domain_node = get_xml_node(xml_node, Elements.DOMAIN)
        domain = Zone_Domain.from_xml_node(domain_node)
        return cls(zone_id, name, domain)

    def _key(self):
        return self.id, self.name

    def __repr__(self):
        return "Zone({id},{name},{comment},{domain})".format(**self.__dict__)


class Zone_Entries_List(XML_List):
    def __init__(self, zone_entries):
        """
        :type zone_entries: list[Zone_Entry]
        """
        self.zone_entries = zone_entries
        super().__init__(Elements.ZONE_ENTRIES, zone_entries)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        zone_entries = []
        for zone_entry_node in xml_node.iter(Elements.ZONE_ENTRY):
            zone_entries.append(Zone_Entry.from_xml_node(zone_entry_node))
        return cls(zone_entries)


class Zone_Entry(XML_Object_Base, Comparable, IPNetworkMixin):
    def __init__(self, item_id, comment, ip, _, netmask, zone_id):
        """

        :type netmask: str|int
        :type comment: str|None
        :type item_id: int|None
        :type zone_id: int|None
        """
        self.id = item_id
        self.comment = comment
        self.ip = ip
        self.netmask = netmask
        self.zoneId = zone_id
        IPNetworkMixin.__init__(self)
        super().__init__(Elements.ZONE_ENTRY)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        item_id = get_xml_int_value(xml_node, Elements.ID)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        ip = get_xml_text_value(xml_node, Elements.IP)
        netmask = get_xml_text_value(xml_node, Elements.NETMASK)
        zone_id = get_xml_int_value(xml_node, Elements.ZONEID)
        return cls(item_id, comment, ip, None, netmask, zone_id)

    def _get_ip_network(self):
        if not self.netmask:
            ip_network = netaddr.IPNetwork(self.ip)
        else:
            ip_network = netaddr.IPNetwork(str(self.ip) + "/" + str(self.netmask))
        return ip_network

    def _key(self):
        return self.id, self.zoneId

    def __repr__(self):
        return "Zone_Entry({id},{comment},{ip},{netmask},{zoneId})".format(**self.__dict__)

    def __str__(self):
        return "{}/{}".format(self.ip, self.netmask)


class ZoneDescendantsList(XML_List):
    def __init__(self, zones):
        self.zones = zones
        super().__init__(Elements.ZONES, zones)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        zones = []
        for zone_node in xml_node.iterfind(Elements.ZONE):
            zones.append(ZoneDescendants.from_xml_node(zone_node))
        return cls(zones)


class ZoneDescendants(XML_Object_Base, Comparable):
    def __init__(self, zone_id, name, zones):
        self.id = zone_id
        self.name = name
        self.zones = zones
        super().__init__(Elements.ZONE)

    @classmethod
    def from_xml_node(cls, xml_node):
        zone_id = get_xml_int_value(xml_node, Elements.ID)
        name = get_xml_text_value(xml_node, Elements.NAME)
        zones = XML_List.from_xml_node_by_tags(xml_node, Elements.ZONES, Elements.ZONE, ZoneDescendants)
        return cls(zone_id, name, zones)

    def _key(self):
        return self.id, self.name

    def __repr__(self):
        return "Zone({id},{name})".format(**self.__dict__)






import shlex
import subprocess
import re
import xml.etree.ElementTree as ET

from .xml import get_xml_node, get_xml_text_value

TSS_EXECUTABLE_PATH = "/usr/sbin/tss"
ST_CONF_XML_PATH = "/etc/sysconfig/stconf.xml"
TOS_VERSION_REGEX = re.compile(
    r"Tufin (?:Security|Orchestration) Suite version: (?P<tos_version_major>\d+)\.(?P<tos_version_minor>\d+)")


def get_st_conf():
    with open(ST_CONF_XML_PATH) as xml_file:
        xml_data = xml_file.read()
    xml_node = ET.fromstring(xml_data)
    return xml_node


def is_ha_server_status_active():
    active_status = ("active", "active_after_failover")
    standby_status = ("standby", "standby_after_failover")
    st_conf_xml_node = get_st_conf()
    ha_conf_node = get_xml_node(st_conf_xml_node, "ha_configuration")
    ha_type = get_xml_text_value(ha_conf_node, "ha_type")
    if ha_type in active_status:
        return True
    elif ha_type == standby_status:
        return False
    else:
        raise ValueError("Unknown HA status '{}'.".format(ha_type))

def get_server_type():
    st_conf_xml_node = get_st_conf()
    server_type = get_xml_text_value(st_conf_xml_node, "Server_Type")
    return server_type


def get_tos_version():
    """
    Get the version reported by "tss version" (e.g. (14,3))
    :return: The major and minor version number for TOS/TSS.
    :rtype: tuple[int,int]
    """
    tss_process = subprocess.Popen([TSS_EXECUTABLE_PATH, 'ver'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    tss_ver_output, _ = tss_process.communicate()
    tos_version_match = re.match(TOS_VERSION_REGEX, tss_ver_output)
    tos_version_major = int(tos_version_match.group("tos_version_major"))
    tos_version_minor = int(tos_version_match.group("tos_version_minor"))
    return tos_version_major, tos_version_minor


def create_backup():
    """
    Create a backup of the local TOS installation.
    :return: The name of the created backup file.
    :rtype: str
    """
    backup_file_name_regex = r"^Compressing and saving Tufin Orchestration Suite backup file to: '?(.*(tgz|zip))"
    print("Creating backup file.")

    export_path_command = 'export PATH="${PATH}:/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin"'
    backup_command = "{} {}".format(TSS_EXECUTABLE_PATH, "backup")
    command_string = ';'.join((export_path_command, backup_command))

    try:
        output = subprocess.check_output(command_string, shell=True).decode()
    except subprocess.CalledProcessError:
        raise IOError("Could not create backup file.")
    backup_file_name = re.search(backup_file_name_regex, output, re.MULTILINE).group(1)
    return backup_file_name
